#ifndef TOPO_MAPPING_TOPOLOGY_H
#define TOPO_MAPPING_TOPOLOGY_H

#include <iostream>
#include <string>
#include <vector>
#include <list>
#include <fstream>
#include <iomanip> //格式化输出的头文件，注意这里不要加.h
//#include <algorithm>

#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <dirent.h>
#include <time.h>

#include <opencv2/opencv.hpp>
#include <boost/concept_check.hpp>
#include <boost/bind.hpp>

#include <nav_msgs/OccupancyGrid.h>		//ROS中的占有图消息格式

#include "topo_mapping/tool.h"

namespace topo_mapping
{
	//拓扑结点信息结构体
	typedef struct VertexNodeInfo
	{
		cv::Point2f pose;		//拓扑结点在全局坐标系map下的坐标
		cv::Point center;
		char rgb_label;		//邻域颜色标识号
// 		int color_h;
// 		int color_s;
// 		int color_v;
		//int radius;			//邻域最大半径
		cv::Point2f axes_size;	//椭圆的短半轴(X)和长半轴(Y)
		//std::vector<cv::Point> points;  //区域包含的所有点
		std::string intention;	//拓扑结点语义，含义或者意图信息
	}VertexNodeInfo;
	typedef struct Edge
	{
		int u;
		int v;
		double weight; 
	}Edge;

	/* 邻接表结构 */
	//边表节点结构，一个adjvex用来存储邻接点的位置，一个nweight用来保存权值，
	typedef struct EdgeNode
	{
		int adjvex;           					//该边所指向的顶点的位置
		double weight;    					//用于存储该边的权值，对于非网图可以不需要 
		bool status;						//表示存在状态
	}EdgeNode;
	
	//顶点节点结构，一个data用来存储数据，一个firstedge是用来指向边表的第一个节点
	typedef struct VertexNode
	{
		int index;					//结点序号，从1开始标号
		int edges;					//边表个数
		struct VertexNodeInfo data;
		std::vector<EdgeNode> adjEdges;//边表容器
		//struct EdgeNode *firstedge;/* 边表头指针 */
	}VertexNode;
	
	class topology
	{
	public:
		topology(double res, int vertexs_res = 100, int vertexs_size = 200, bool debug = false);
		~topology();
	
		/*设置操作，直接设置某些属性*/
		bool setVertexIndex(const int index);
		bool setTopoMapLabel(const cv::Point pixel_pos, const char label);		//设置指定像素点的颜色
		void setAllEdgeInvalid();
		
		/*获取操作，获取某些属性的具体情况*/
		int vertex_size();									//获得顶点个数
		int edge_size();									//获得边表个数
		double getResolution(bool isIntopo = true);								//获得地图分辨率
		cv::Point2f getOrignposeInmap();
		double** getAdjMatrix();
		nav_msgs::OccupancyGrid* getTopoMap();
		char getOccValue(const cv::Point point);
		char getTopoLabel(const cv::Point point);
		VertexNode* getVertexNode(const int locate);
		void getAllEdges(std::vector< Edge >& allEdges);
		VertexNode* getLastVertexInfo();
		void resetVertexIter();
		std::vector<VertexNode>::iterator getVertexIter();
		bool vertexIterIsEnd();
		bool vertexItertemp();
		std::vector<VertexNode>* getVertexsList();
		
		/*更新操作，更新某些属性值*/
		bool updateAdjMatrix();							//更新邻接矩阵信息
		bool updateGridMap(const nav_msgs::OccupancyGridConstPtr& map);
		bool updateTopoMap(const nav_msgs::OccupancyGridConstPtr& map, bool needdata = false);
		void updateMapTime();
		bool isGridMapEmpty();
		bool isTopoMapEmpty();
		
		/*转换操作，对某些特征或者属性进行转换*/
		cv::Point map2pixelCoord(const cv::Point2f mapCoord, bool isIntopo = false);		//将map坐标转换为像素点坐标
		cv::Point2f pixel2mapCoord(const cv::Point pixelCoord, bool isIntopo = false);		//像素坐标，往右方向为x轴正方向，往下方向为y轴正方向
		int pixel2linearIndex(const cv::Point pixelCoord, bool isIntopo = false);			//将像素坐标转换成数组线性序列号，因为地图数据是按行扫描得到的一维数组
		
		/*使用操作，进行插入、删除等操作*/
		int insertAVertex(const VertexNodeInfo msg, char center_label = 127);		//插入一个顶点，返回该顶点使用的序号
		bool addAnEdge(const int vertex2, int vertex1 = 0);
		bool deleteAnEdge(const int vertex2, int vertex1 = 0);					//删除一条边表
		bool deleteAVertex(const int vertex_index, bool deletelabel = false);		//删除一个顶点
		int deleteFalseEdges();								//与setAllEdgeInvalid函数配合使用
		double computeOccupiedRate();
		
		void showTopoMap_graph();
		void showVertexData();
		bool saveTopoMap2TXT(std::string path);
		
		bool vertex_id_compare(VertexNode a, VertexNode b);
		
	private:
		
		nav_msgs::OccupancyGrid GridMap_temp;		//占用栅格消息
		nav_msgs::OccupancyGrid TopoMap_temp;		//拓扑区域消息
		
		int Vertex_id;				//目前状态下的拓扑节点序号
		int Vertex_maxsize;		//拓扑节点最大个数
		bool debug_flag;
		
		/* 邻接表结构 */
		std::vector<VertexNode> VertexsList;
		std::list<int> AbandonedVertexs;
		int VertexesNum;								/* 图中当前顶点数 */
		int EdgesNum; 								/* 图中当前边数 */
		std::vector<VertexNode>::iterator VertexIter;		//拓扑顶点迭代器
		std::vector<VertexNode>::iterator VertexIter_t;		//拓扑顶点迭代器临时保存位置
		double Res_topo;				//地图分辨率
		cv::Point2f Orign_topo;			//地图左上角像素点(0, 0)的map坐标
		
		/* 邻接矩阵结构 */
		int* AdjMatrixVertexs;
		double** AdjMatrixEdges;
		int AdjMatrix_vertexs;

		//栅格地图参数，用于坐标转换
		double Res_grid;				//地图分辨率
		cv::Point2f Orign_grid;			//地图左上角像素点(0, 0)的map坐标
		
	};
}
#endif