#include "topo_mapping/console.h"

namespace topo_mapping
{

	console::console()
		:console_nh("~")
	{
		console_nh.getParam("filepath", MapFile_path);
		console_nh.getParam("vertexs_max_size", VertexsQuantity_max);
		console_nh.getParam("topomap_res", Res_topo);
		//console_nh.getParam("cascade_times", cascade_times);
		
		//获取终端上键盘输入的参数配置
		tcgetattr(Terminal_kfd, &TerminalConfig_ori);							// 得到终端的初始配置并保存
		memcpy(&TerminalConfig_imp, &TerminalConfig_ori, sizeof(struct termios));
		TerminalConfig_imp.c_lflag &=~ (ICANON | ECHO);
		TerminalConfig_imp.c_cc[VEOL] = 1;
		TerminalConfig_imp.c_cc[VEOF] = 2;
		Terminal_ufd.fd = Terminal_kfd;
		Terminal_ufd.events = POLLIN;			//检测可读状态
		
		ROS_INFO("%s", USAGE);
	}
	console::~console()
	{
		tcsetattr(Terminal_kfd, TCSANOW, &TerminalConfig_ori);  
		TopoMap->saveTopoMap2TXT(MapFile_path);
		
		delete MappBuilder;
		delete TopoMap;
	}
	void console::runPreparation()
	{
		gridmap_sub = console_nh.subscribe("/map", 1, &console::gridmapCallback, this);
		path_sub = console_nh.subscribe("/pose_path", 1, &console::robotpathCallback, this);
		vertexMarkers_pub = console_nh.advertise<visualization_msgs::MarkerArray>("/semi_navigation/VertexMarkers", 1);  // map
		vertexConnectors_pub = console_nh.advertise<visualization_msgs::MarkerArray>("/semi_navigation/connection_relation", 1); 
		vertexs_ser = console_nh.advertiseService("/dynamic_topomap", &console::vertexs_request, this);
		
		// 确保发布标记时，已经打开rviz进行显示
		while (vertexMarkers_pub.getNumSubscribers() < 1 && ros::ok())
		{
			ROS_WARN_ONCE("Please create a subscriber to the marker or launch rviz ! ");
			ros::Duration(0.5);
		}

		//创建一个拓扑地图类，用于保存拓扑地图信息
		TopoMap = new topology(Res_topo, VertexsQuantity_max, VertexsQuantity_max*2, false);
		
		
		//创建一个地图构建类，用于构建拓扑地图
		MappBuilder = new mapping(TopoMap, &console_nh, false);
		
		
		// 检查栅格地图数据
		while (TopoMap->isGridMapEmpty() && ros::ok())
		{
			ros::spinOnce();
			ROS_WARN_ONCE("Wait for gridmap data ...... ! ");
			ros::Duration(0.5);
		}
		
	}
	void console::gridmapCallback(const nav_msgs::OccupancyGridConstPtr& map)
	{
		ROS_INFO_THROTTLE(10, "Received a %d X %d gridmap @ %.3f m/pix",
				map->info.width,
				map->info.height,
				map->info.resolution);				//代码运行5次输出一次message． 
		
		TopoMap->updateGridMap(map);
		
		static bool map_init = true;
		if(map_init == false)
		{
			TopoMap->updateTopoMap(map, true);
			map_init = true;
		}else
			TopoMap->updateTopoMap(map, false);
	}
	void console::robotpathCallback(const nav_msgs::PathConstPtr& path)
	{
		int pose_size = path->poses.size();
		if(pose_size > robot_trajectory.poses.size())
		{
			ROS_INFO("Received a robot trajectory with %d poses", pose_size);
			robot_trajectory.header = path->header;
			robot_trajectory.poses = path->poses;
		}
	}
	bool console::vertexs_request(vertex::Request& req, vertex::Response& res)
	{
		if((TopoMap->vertex_size()) < 1)
			return false;
		
		//判断收到的指令是否正确
		if(req.order == 1)
		{
			TopoMap->resetVertexIter();
			std::vector<VertexNode>::iterator iter_v;	
			do{
				iter_v = TopoMap->getVertexIter();
				res.vertex.push_back((*iter_v).data.center.x);
				res.vertex.push_back((*iter_v).data.center.y);
				res.vertex.push_back((*iter_v).data.axes_size.x);
				res.vertex.push_back((*iter_v).data.axes_size.y);
			}while(TopoMap->vertexIterIsEnd() == false);
		}else if(req.order == 2)
		{
			double temp;
			console_nh.getParam("obstacle_ratio", temp);
			res.vertex.push_back(temp);
			console_nh.getParam("add_ratio", temp);
			res.vertex.push_back(temp);
			console_nh.getParam("pass_ratio", temp);
			res.vertex.push_back(temp);
			console_nh.getParam("radius_weight1", temp);
			res.vertex.push_back(temp);
		}

		return true;
	}
	int console::KeyboardDetect(int blockingTime)
	{
		char key = 0;
		int key_value = 0;
		
		tcsetattr(Terminal_kfd, TCSANOW, &TerminalConfig_imp);		//设置参数
		if ((poll(&Terminal_ufd, 1, blockingTime)) >0)					//阻塞时间（毫秒）
		{
			if(read(Terminal_kfd, &key, 1) > 0) 						//读取键盘键值
			{	
				//键值查询
				switch(key)
				{
					case KEYCODE_W: key_value = 1; break;
					case KEYCODE_A: key_value = 2; break;
					case KEYCODE_S: key_value = 3; break;
					case KEYCODE_D: key_value = 4; break;
					case KEYCODE_W_CAP: key_value = 5; break;
					case KEYCODE_A_CAP: key_value = 6; break;
					case KEYCODE_S_CAP: key_value = 7; break;
					case KEYCODE_D_CAP: key_value = 8; break;
					default:  key_value = 0 ;
				}
			}
		}
		tcsetattr(Terminal_kfd, TCSANOW, &TerminalConfig_ori);  //恢复终端键盘输入的初始状态
		return key_value;
	}
	bool console::KeyEventProcessing()
	{
		int key_value = KeyboardDetect();
		int edge_a, edge_b;
		
		
		if(key_value == 0)
			return false;
		
		if(key_value == 1)
		{//w
			std::cout<<"console: The details of each topology vertex will be displayed immediately below !"<<std::endl;
			TopoMap->showVertexData();
			//MappBuilder->establishATopoVertex();
			return true;
		}
		if(key_value == 2)
		{//a
			std::cout<<"console: The detailed connection of the topology will be displayed immediately below !"<<std::endl;
			TopoMap->showTopoMap_graph();
			return true;
		}
		if(key_value == 3)
		{//s
			std::cout<<"console: The edges of the vertices will be manipulated below !"<<std::endl;
			int edge_a, edge_b, flag;
			std::cout<<" 请输入要操作的边顶点号(以空格形式隔开)：";
			std::cin>>edge_a>>edge_b;
			std::cout<<" 请问是要删除还是添加(删除输入1,添加输入2)：";
			std::cin>>flag;
			if(flag == 1)
				TopoMap->deleteAnEdge(edge_a, edge_b);
			else
				TopoMap->addAnEdge(edge_a, edge_b);
			return true;
		}
		if(key_value == 4)
		{//d
			std::cout<<"console: The minimum spanning tree form adjustment will be made to the topology below !"<<std::endl;
			 MappBuilder->adjustmentTopoGraph();
			return true;
		}
		if(key_value == 5)
		{//W
			std::cout<<"console: The metric map will be topologically segmented to form a topological map !"<<std::endl;

			//获得机器人运动轨迹
			robot_trajectory.poses.clear();
			while(robot_trajectory.poses.size() == 0)
				ros::spinOnce();
			
			int pose_size = robot_trajectory.poses.size();
			cv::Point2f robot_pose;
			for(int i = 0 ; i < pose_size ; i++)
			{
				robot_pose.x = robot_trajectory.poses[i].pose.position.x;
				robot_pose.y = robot_trajectory.poses[i].pose.position.y;
				MappBuilder->NeighborhoodSegmentation(robot_pose);
				
				showVertexMarks_rviz();
				showVertexConnectors_rviz();
			}
			return true;
		}
		if(key_value == 6)
		{//A
			std::cout<<"console: Secondary growth of the topological map !"<<std::endl;
			std::cout<<"重新生长前占有率为:"<<(TopoMap->computeOccupiedRate())<<std::endl;
			bool flag = true;
			TopoMap->resetVertexIter();
			while(flag && ros::ok())
			{
				flag = MappBuilder->RegionalRegrowth();
				
				showVertexMarks_rviz();
				showVertexConnectors_rviz();
			}
			std::cout<<"区域重新生长完毕！"<<std::endl;
			std::cout<<"重新生长后占有率为:"<<(TopoMap->computeOccupiedRate())<<std::endl;
			return true;
		}
		if(key_value == 7)
		{//S					
			std::cout<<"console: Convergence of similar topological regions !"<<std::endl;
			static double merger_times = 3;
			static int radius_max = round(MappBuilder->Radius_max / (TopoMap->getResolution()));
			static double area_max = M_PI * pow(radius_max, 2);
			for(int id = 0 ; id < TopoMap->vertex_size() ; id++)
			{
				MappBuilder->RegionalMerger(id, area_max/merger_times);		
				
				showVertexMarks_rviz();
				showVertexConnectors_rviz();
			}
			std::cout<<"融合过程结束 ! "<<std::endl;
			merger_times--;
			if(merger_times <= 0)
			{
			  merger_times = 3;
			  std::cout<<"融合过程全部结束 ! "<<std::endl;
			}
			return true;
		}
		if(key_value == 8)
		{//D
			
			
// 			static double ratio = 0;
// 			ratio = ratio + 0.025;
// 			MappBuilder->RegionalMerger2(ratio);
			return true;
		}
	}
	void console::showVertexMarks_rviz()
	{
		visualization_msgs::MarkerArray ma;
		visualization_msgs::Marker marker;
		
		marker.header.frame_id = "/map";
		marker.id = 0;
		marker.ns = "VertexNodes";
		marker.type = visualization_msgs::Marker::CYLINDER;//圆柱
		marker.action = visualization_msgs::Marker::ADD;
		marker.lifetime = ros::Duration(0.4);
		
		marker.color.r = 0/255.0;//黑色
		marker.color.g = 0/255.0;
		marker.color.b = 0/255.0;
		marker.color.a = 0.6;//设置标记颜色，确保alpha（透明度）值为0
		
		TopoMap->resetVertexIter();
		std::vector<VertexNode>::iterator iter_v;	
		do{
			if(TopoMap->vertex_size() == 0)
				break;
			iter_v = TopoMap->getVertexIter();
			marker.id++;
			
			//填入标记形状参数
			tf::Quaternion q;
			q.setRPY(0, 0, 90);
			
			marker.type = visualization_msgs::Marker::CYLINDER;
			marker.pose.position.x = ((*iter_v).data.pose.x);
			marker.pose.position.y = ((*iter_v).data.pose.y);
			marker.pose.position.z = 0;
			marker.pose.orientation.x = q.x();
			marker.pose.orientation.y = q.y();
			marker.pose.orientation.z = q.z();
			marker.pose.orientation.w = q.w();
			marker.scale.x = 0.3;
			marker.scale.y = 0.3;
			marker.scale.z = 0.1;  		// Z can't be 0, limitation of ROS
			marker.color.a = 0.6;	
			ma.markers.push_back(marker);
			
			//写入标记数字
			marker.id++;
			marker.type = visualization_msgs::Marker::TEXT_VIEW_FACING;
			marker.text = std::to_string((*iter_v).index);
			marker.pose.position.x = (*iter_v).data.pose.x + 0.3;
			marker.pose.position.y = (*iter_v).data.pose.y + 0.3;
			marker.pose.position.z = 0.1;
			marker.color.a = 1.0;	
			marker.scale.x = 0.5;
			marker.scale.y = 0.5;
			marker.scale.z = 0.5;
			ma.markers.push_back(marker);
		}while(TopoMap->vertexIterIsEnd() == false);
		vertexMarkers_pub.publish(ma);
	}
	void console::showVertexConnectors_rviz()
	{
		float line_height = 0.1;
		int points_num = 1;
		
		//visualization of the data association
		visualization_msgs::MarkerArray points_associate;
		visualization_msgs::Marker line_strip;
		
		line_strip.header.frame_id = "/map";
		line_strip.header.stamp = ros::Time::now();

		line_strip.id = 0;
		line_strip.ns = "connection_relationship";
		line_strip.type = visualization_msgs::Marker::LINE_STRIP;
		line_strip.action = visualization_msgs::Marker::ADD;
		
		line_strip.lifetime = ros::Duration(0.4);
		line_strip.pose.position.z = 0.1;
		line_strip.pose.orientation.x = 0.0;
		line_strip.pose.orientation.y = 0.0;
		line_strip.pose.orientation.z = 0.0;
		line_strip.pose.orientation.w = 1.0;
		line_strip.scale.x = 0.1;  // line uses only x component
		line_strip.scale.y = 0.1;
		line_strip.scale.z = 0.1;
		line_strip.color.a = 1.0;
		line_strip.color.r = 1.0;
		line_strip.color.g = 1.0;
		line_strip.color.b = 1.0;
		
		geometry_msgs::Point start_point, end_point;
		start_point.z = line_height;
		end_point.z = line_height;

		std::vector< Edge > allEdgedges;
		TopoMap->getAllEdges(allEdgedges);
		Edge edge_temp;
		for(int i = 0 ; i < allEdgedges.size() ; i++)
		{
			edge_temp = allEdgedges[i];
			start_point.x = (TopoMap->getVertexNode(edge_temp.u))->data.pose.x;
			start_point.y = (TopoMap->getVertexNode(edge_temp.u))->data.pose.y;
			end_point.x = (TopoMap->getVertexNode(edge_temp.v))->data.pose.x;
			end_point.y = (TopoMap->getVertexNode(edge_temp.v))->data.pose.y;
			
			line_strip.points.clear();
			line_strip.points.push_back(start_point);
			line_strip.points.push_back(end_point);
			points_associate.markers.push_back(line_strip);
			line_strip.id++;
		}
		vertexConnectors_pub.publish(points_associate);
	}
	bool console::mainLoop()
	{
		ros::Rate r(60);
		while(ros::ok())
		{
			ros::spinOnce();
			KeyEventProcessing();

			showVertexMarks_rviz();
			showVertexConnectors_rviz();
			
			r.sleep();
		}
	}







}
