#include "topo_mapping/mapping.h"

namespace topo_mapping
{
	mapping::mapping(topology* mTopoMap, ros::NodeHandle* nh, bool debug_)
		:maTopoMap(mTopoMap), private_nh(nh), debug_flag(debug_)
	{
		private_nh->getParam("max_radius", Radius_max);
		private_nh->getParam("min_radius", Radius_min);
		private_nh->getParam("expansion_size", Expansion_size);
		private_nh->getParam("obstacle_ratio", ObstacleRatio);
		private_nh->getParam("add_ratio", AddRatio);
		private_nh->getParam("pass_ratio", PassRatio);
		private_nh->getParam("radius_weight1", Radius_w1);
		
		AdjacentVertex = 0;
		Pen_color = -120;
		
		topomap_pub = private_nh->advertise<nav_msgs::OccupancyGrid>("/topomap", 1, true);
		samples_pub = private_nh->advertise<visualization_msgs::MarkerArray>("/semi_navigation/samples", 1); 
		topo_pointMarkers = private_nh->advertise<visualization_msgs::MarkerArray>("/semi_navigation/regions", 1); 
	}
	mapping::~mapping()
	{

	}
	bool mapping::regionsMarkers_pub(int marker_id, std::vector<cv::Point> points)
	{
		visualization_msgs::MarkerArray ma;
		visualization_msgs::Marker marker;
		
		marker.header.frame_id = "map";
		marker.id = 0;
		marker.ns = "regional_points";
		marker.type = visualization_msgs::Marker::CUBE;//圆柱
		marker.action = visualization_msgs::Marker::ADD;
		marker.lifetime = ros::Duration(0);
		
		marker.color.r = 0/255.0;//蓝色
		marker.color.g = 0/255.0;
		marker.color.b = 255/255.0;
		marker.color.a = 0.6;//设置标记颜色，确保alpha（透明度）值为0
		
		tf::Quaternion q;
		q.setRPY(0, 0, 0);
		marker.pose.position.z = 0;
		marker.pose.orientation.x = q.x();
		marker.pose.orientation.y = q.y();
		marker.pose.orientation.z = q.z();
		marker.pose.orientation.w = q.w();
		
		marker.scale.x = maTopoMap->getResolution();
		marker.scale.y = maTopoMap->getResolution();
		marker.scale.z = maTopoMap->getResolution();  		// Z can't be 0, limitation of ROS
		
		for(int i = 0 ; i < points.size() ; i++)
		{
			marker.id++;
			
			//填入标记形状参数
 			cv::Point2f point = maTopoMap->pixel2mapCoord(points[i], true);
			marker.pose.position.x = point.x;
			marker.pose.position.y = point.y;

			ma.markers.push_back(marker);
		}
		topo_pointMarkers.publish(ma);
	}
	bool mapping::getRobotPose(cv::Point2f& pose)
	{
		bool got_tf = false;
		tf::StampedTransform world_pose;
		tf::TransformListener listener_;
		
		// Update the pose of the robot
		while(!got_tf && ros::ok()){
			try{
				listener_.lookupTransform("/map","/base_link", ros::Time(0), world_pose);
				got_tf = true;
			}
				catch (tf::TransformException ex) {
				ROS_INFO_STREAM_ONCE("Wait for tf: from /map to /base_link !"); 		//在循环中让信息只输出一次 
			} 
			ros::Duration(0.05).sleep();
		}
		if(got_tf)
		{
			pose.x =  world_pose.getOrigin().getX();
			pose.y = world_pose.getOrigin().getY();
			return true;
		}else{
			ROS_ERROR("ERROR to lookup transform !"); 
			return false;
		}
	}
	bool mapping::establishATopoVertex(const cv::Point2f axesSize, cv::Point pixelCenter, char colorLabel)
	{
		VertexNodeInfo vertex_msg;
		cv::Point2f robot_pose;
		
		if((pixelCenter.x == 0)&&(pixelCenter.y == 0))//如果没有指定顶点，则获取当前机器人位置点
		{
			if(getRobotPose(robot_pose))
			{
				vertex_msg.center = maTopoMap->map2pixelCoord(robot_pose, true);
			}else{
				ROS_ERROR("ERROR to lookup transform !"); 
				return false;
			}
		}else{
			vertex_msg.center = pixelCenter;
		}
		
// 		std::string temp;
// 		std::cout<<"请输入该节点的含义，语义或者意图：";
// 		std::cin>>temp;
// 		if(temp == "q" || temp == "Q")
// 		{
// 			std::cout<<"接收到退出指令，即将退出......"<<std::endl;
// 			ros::shutdown();
// 		}else vertex_msg.intention = temp;
		vertex_msg.intention = "  ";
		if(colorLabel == 0)
			vertex_msg.rgb_label = Pen_color;
		else 
			vertex_msg.rgb_label = colorLabel;

		vertex_msg.axes_size = axesSize;
		int index = maTopoMap->insertAVertex(vertex_msg);	//插入拓扑节点
		addAdjEdges(index);					//给当前拓扑节点添加边表节点，确定邻边关系，注意至多四条邻边
		
		return true;
	}
	bool mapping::statisticsAdjEdges(std::vector< int >& adjEdges, const int locate)
	{
		VertexNode* region_center = maTopoMap->getVertexNode(locate);
		cv::Point center_pixel = region_center->data.center;
		
		//统计区域最外围合格的栅格点
		std::vector<cv::Point> spiral_points;
		std::vector<cv::Point>::iterator iter_sp;
		std::vector<cv::Point> points;
		
		int scan_range = round( (region_center->data.axes_size.x)*1.2 );							//设置起始半径
		pixelSpiralIterator(center_pixel, scan_range, points);									//环形扫描
		spiral_points.insert(spiral_points.end(), points.begin(), points.end());
		scan_range = round( (region_center->data.axes_size.y)*1.2 );								//设置起始半径
		pixelSpiralIterator(center_pixel, scan_range, points);									//环形扫描
		spiral_points.insert(spiral_points.end(), points.begin(), points.end());
		for(iter_sp = spiral_points.begin() ; iter_sp != spiral_points.end() ; )
		{
			if(checkPointISQualified(*iter_sp, 0, 0) == false)
				iter_sp = spiral_points.erase(iter_sp);
			else
				++iter_sp;
		}

		//统计栅格点对应的颜色标签
		std::vector<char> labels;
		std::vector<char>::iterator iter_label;
		labels.clear();
		for(int i = 0 ; i < spiral_points.size() ; i++)
		{
			char label = maTopoMap->getTopoLabel( spiral_points[i] );
			iter_label = std::find(labels.begin(), labels.end(), label);
			if(iter_label == labels.end())//未找到重复标签
				labels.push_back(label);
		}
		if(labels.empty())
			return false;
		
		//统计扫描范围内的拓扑顶点
		adjEdges.clear();
		maTopoMap->resetVertexIter();
		std::vector<VertexNode>::iterator iter_v;
		do{
			iter_v = maTopoMap->getVertexIter();
			if((*iter_v).index != locate)			//排除自己
			{
				if(tool::getNorm(region_center->data.pose, (*iter_v).data.pose) < (Radius_max*2.3))
					adjEdges.push_back((*iter_v).index);
			}
		}while(maTopoMap->vertexIterIsEnd() == false);
		
		//判断顶点之间是否可达，中间无障碍物
		std::vector< cv::Point > traversal_points;	//端点连线经过的所有点
		std::vector<int>::iterator iter_adj;
		VertexNode* temp;
		for( iter_adj = adjEdges.begin() ; iter_adj != adjEdges.end() ;  )
		{
			//计算终点栅格坐标
			temp = maTopoMap->getVertexNode(*iter_adj);
			cv::Point end = temp->data.center;
			ImprovedBresenham(center_pixel, end, traversal_points);		//利用Bresenham划线算法计算途径的栅格点坐标
			int t = 0;
			for( t = 0 ; t < traversal_points.size() ; t++)
			{	
				if(checkPointISQualified(traversal_points[t], region_center->data.rgb_label, temp->data.rgb_label) == false)
				{
					iter_adj = adjEdges.erase( iter_adj ); 
					break;
				}
			}
			if(t == traversal_points.size())//未检测到障碍物
				++iter_adj;    						//指向下一个位置
		}
		if(adjEdges.empty())
			return false;
		
		//最终判断是否符合拓扑邻域颜色标识
		for( iter_adj = adjEdges.begin() ; iter_adj != adjEdges.end() ;  )
		{
			char label = (maTopoMap->getVertexNode(*iter_adj))->data.rgb_label;
			iter_label = std::find(labels.begin(), labels.end(), label);
			if(iter_label == labels.end())
				iter_adj = adjEdges.erase( iter_adj );  
			else
				++iter_adj;
		}
	}
	bool mapping::statisticsAdjVertexs(std::vector< int >& adjVertexs, const cv::Point2f locate)
	{
		//首先计算欧几里得距离合适的拓扑顶点
		adjVertexs.clear();
		maTopoMap->resetVertexIter();
		std::vector<VertexNode>::iterator iter_v;
		do{
			iter_v = maTopoMap->getVertexIter();
			double dist = tool::getNorm(locate, (*iter_v).data.pose);
			if(dist < (Radius_max*1.3) && dist > Radius_min * 0.8) 
				adjVertexs.push_back((*iter_v).index);
		}while(maTopoMap->vertexIterIsEnd() == false);
		
		//然后判断两个顶点是否可达，中间无障碍物
		cv::Point start, end;					//需要计算的首末端点
		std::vector< cv::Point > traversal_points;	//端点连线经过的所有点
		std::vector<int>::iterator iter;
		start = maTopoMap->map2pixelCoord(locate);
		for( iter = adjVertexs.begin() ; iter != adjVertexs.end() ;  )
		{
			//计算终点栅格坐标
			end = maTopoMap->map2pixelCoord(((maTopoMap->getVertexNode(*iter))->data.pose));
			ImprovedBresenham(start, end, traversal_points);		//利用Bresenham划线算法计算途径的栅格点坐标
			int t = 0;
			for( t = 0 ; t < traversal_points.size() ; t++)
			{
				//获取栅格点的占有率
				if(maTopoMap->getOccValue(traversal_points[t]) == 100)	//有障碍物或者未知区域
				{
					iter = adjVertexs.erase( iter );    				//删除元素，返回值指向已删除元素的下一个位置
					break;
				}
			}
			if(t == traversal_points.size())//未检测到障碍物
				++iter;    						//指向下一个位置
		}
		
	}
	bool mapping::changePenColor()
	{
		if(Pen_color >=120)		//更新邻域颜色标识
			Pen_color = -120;
		else 
			Pen_color = Pen_color + 20;
		if(Pen_color == 0)
			Pen_color = Pen_color + 20;
	}
	bool mapping::addAdjEdges(const int locate)
	{
		std::vector<int> adj_edges;
		statisticsAdjEdges(adj_edges, locate);			//计算相邻顶点，补充边表
// 		cv::Point2f vertex_pose = (maTopoMap->getVertexNode(locate))->data.pose;
// 		statisticsAdjVertexs(adj_edges, vertex_pose);
		
		if(adj_edges.size() >0)
		{//添加所有边表
			
			std::vector<char> edges_label;
			for(int i = 0 ; i < adj_edges.size() ; i++)
			{
				maTopoMap->addAnEdge(locate, adj_edges[i]);							//给当前顶点添加边表
				maTopoMap->addAnEdge(adj_edges[i], locate);
				edges_label.push_back(maTopoMap->getVertexNode(adj_edges[i])->data.rgb_label);
				
			}	
			
			VertexNode* vertex_map = maTopoMap->getVertexNode(locate);
			char label_ori = vertex_map->data.rgb_label;		//拓扑区域原色
			char label_alter = label_ori;						//拓扑区域欲填色
			std::vector<char>::iterator iter_check;
			iter_check = edges_label.begin();
			while(iter_check != edges_label.end())
			{
			  iter_check = std::find(edges_label.begin(), edges_label.end(), label_alter);
			  if(iter_check != edges_label.end())//发现重复色
			  {
				  if(Pen_color == label_alter)		//检查画笔的当前颜色是否可用，是否和周围颜色重复，若是则改
					  changePenColor();
				  label_alter = Pen_color;
			  }
			}
			if(label_alter != label_ori)
			{
				if(debug_flag)
					std::cout<<"发生颜色重复，对颜色标识进行修改:"<<(int)(label_ori)<<">>"<<(int)(label_alter)<<std::endl;
				//对颜色标识进行修改
				std::vector< cv::Point > traversal_points;	//环形遍历得到的像素点(栅格点)
				cv::Point center = vertex_map->data.center;
				int range_max = round((vertex_map->data.axes_size.y) * 1.4);
				int radius = 1;
				while(ros::ok() &&  radius < range_max)
				{
					pixelSpiralIterator(center, radius, traversal_points);			//环形遍历像素点，一次获得一个环形数据
					radius++;
					for(int i = 0 ; i < traversal_points.size() ; i++)
					{
						if(maTopoMap->getTopoLabel(traversal_points[i]) == label_ori)
							maTopoMap->setTopoMapLabel(traversal_points[i], label_alter);
					}
					
					//发布实时消息
					maTopoMap->updateMapTime();
					topomap_pub.publish( *(maTopoMap->getTopoMap()) );
				}
				vertex_map->data.rgb_label = label_alter;
			}
		}
		if(debug_flag)
			std::cout<<"mapping: 已经完成"<<(adj_edges.size())<<"条边表的插入！"<<std::endl;
	}
	bool mapping::adjustmentTopoGraph()
	{
		std::vector< Edge > miniSpanTree;
		int vertex_num = (maTopoMap->getLastVertexInfo())->index + 1;
		
		std::vector<Edge> allEdges;
		maTopoMap->getAllEdges(allEdges);
 		MiniSpanTree_Kruskal(vertex_num, allEdges, miniSpanTree);
		
		//FIXME 这里的某个环节存在问题，暂未查出
// 		std::cout<<"adjustmentTopoGraph <1>"<<std::endl;
// 		if(maTopoMap->updateAdjMatrix())
// 		{
// 			double** adj_matrix = maTopoMap->getAdjMatrix();
// 			MiniSpanTree_Prim(vertex_num, adj_matrix, miniSpanTree);
// 		}else{
// 			std::cerr<<"The adj_matrix is error from updateAdjMatrix() !"<<std::endl;
// 			return false;
// 		}
// 		std::cout<<"adjustmentTopoGraph <2>"<<(miniSpanTree.size())<<std::endl;
// 		for(int i = 0 ; i < miniSpanTree.size() ; i++)
// 		  std::cout<<(miniSpanTree[i].u)<<" "<<(miniSpanTree[i].v)<<std::endl;
		
		VertexNode* vertex_temp;
		maTopoMap->setAllEdgeInvalid();										//将所有边置为无效状态
		for(int i = 0 ; i < miniSpanTree.size() ; i++ )						//将最小生成树所包含的边恢复为有效状态
		{
			vertex_temp = maTopoMap->getVertexNode(miniSpanTree[i].u);
			for(int j = 0 ; j < vertex_temp->adjEdges.size() ; j++)
			{
				if(vertex_temp->adjEdges[j].adjvex == miniSpanTree[i].v)
				{
					vertex_temp->adjEdges[j].status = true;
					break;
				}
			}
			
			vertex_temp = maTopoMap->getVertexNode(miniSpanTree[i].v);
			for(int j = 0 ; j < vertex_temp->adjEdges.size() ; j++)
			{
				if(vertex_temp->adjEdges[j].adjvex == miniSpanTree[i].u)
				{
					vertex_temp->adjEdges[j].status = true;
					break;
				}
			}
		}
		
		//删除无效边
		int delete_num = maTopoMap->deleteFalseEdges();
		if(debug_flag)
			std::cout<<"拓扑无向图重调整完毕！"<<std::endl;
	}
	bool mapping::chooseBestGrowthPoint(cv::Point2f& growthPoint)
	{
		//初始生长点选择
		//FIXME 此处方法不完善
		cv::Point2f robot_pose;

		if(getRobotPose(robot_pose))						//获取机器人位置
		{
			if((growthPoint.x != 0)&&(growthPoint.y != 0))	//排除系统初始化状态
			{
				//条件1，确保当前位置处于无拓扑区域属性的地区
				if(maTopoMap->getTopoLabel(maTopoMap->map2pixelCoord(robot_pose, true)) == 0)
				{
					//条件2，确保与最近拓扑顶点有一段距离
					//获取最近拓扑顶点状态
					VertexNode* adjvertex;
					if(AdjacentVertex == 0)		//临近检测状态初始化
						AdjacentVertex =  (maTopoMap->getLastVertexInfo())->index;
					else{//如果存在最近顶点，判断该顶点是否合理，是否需要替换
						double min_dis = 255;
						adjvertex = (maTopoMap->getVertexNode(AdjacentVertex));
						if(adjvertex != NULL)
							min_dis = tool::getNorm(robot_pose, adjvertex->data.pose);
						
						std::vector<int> adj_vertexs;
						statisticsAdjVertexs(adj_vertexs, robot_pose);		//计算相邻顶点，选取最近点判断距离

						for(int i = 0 ; i < adj_vertexs.size() ; i++)
						{
							cv::Point2f temp = (maTopoMap->getVertexNode(adj_vertexs[i]))->data.pose;
							double dis = tool::getNorm(robot_pose, temp);
							if(dis < min_dis)
							{
								AdjacentVertex = adj_vertexs[i];
								min_dis = dis;
							}
						}
					}
					
					//计算与最近拓扑顶点的欧式距离
					adjvertex = (maTopoMap->getVertexNode(AdjacentVertex));
					double dist = (tool::getNorm(robot_pose, adjvertex->data.pose)) - 
								(adjvertex->data.axes_size.x + adjvertex->data.axes_size.y)*0.5*maTopoMap->getResolution(true);
					if( dist <= (Radius_min * 0.6))
						return false;
				}else
					return false;
			}	
			//系统初始化，生长点随机分配
			
			growthPoint = robot_pose;
			if(debug_flag)
				std::cout<<"mapping: 开始进行区域生长......此时临近点为 "<<(AdjacentVertex)<<std::endl;
		}else
			return false;
		
		return true;
	}
	bool mapping::checkBestGrowthPoint(cv::Point2f growthPoint)
	{
		if(maTopoMap->vertex_size() == 0)
			return true;
		
		//条件1，确保当前位置处于无拓扑区域属性的地区
		if(maTopoMap->getTopoLabel(maTopoMap->map2pixelCoord(growthPoint, true)) == 0)
		{
			//条件2，确保与最近拓扑顶点有一段距离
			//获取最近拓扑顶点状态
			VertexNode* adjvertex;
			if(AdjacentVertex == 0)		//临近检测状态初始化
				AdjacentVertex =  (maTopoMap->getLastVertexInfo())->index;
			else{//如果存在最近顶点，判断该顶点是否合理，是否需要替换
				double min_dis = 255;
				adjvertex = (maTopoMap->getVertexNode(AdjacentVertex));
				if(adjvertex != NULL)
					min_dis = tool::getNorm(growthPoint, adjvertex->data.pose);
				
				std::vector<int> adj_vertexs;
				statisticsAdjVertexs(adj_vertexs, growthPoint);		//计算相邻顶点，选取最近点判断距离

				for(int i = 0 ; i < adj_vertexs.size() ; i++)
				{
					cv::Point2f temp = (maTopoMap->getVertexNode(adj_vertexs[i]))->data.pose;
					double dis = tool::getNorm(growthPoint, temp);
					if(dis < min_dis)
					{
						AdjacentVertex = adj_vertexs[i];
						min_dis = dis;
					}
				}
			}
			
			//计算与最近拓扑顶点的欧式距离
			adjvertex = (maTopoMap->getVertexNode(AdjacentVertex));
			double dist = (tool::getNorm(growthPoint, adjvertex->data.pose)) - 
						(adjvertex->data.axes_size.x + adjvertex->data.axes_size.y)*0.5*maTopoMap->getResolution(true);
			if( dist <= (Radius_min * 0.6))
				return false;
		}else
			return false;
		
		if(debug_flag)
			std::cout<<"mapping: 开始进行区域生长......此时临近点为 "<<(AdjacentVertex)<<std::endl;
		
		return true;
	}
	bool mapping::pixelDynamicGrowth(cv::Point& pixelCenter, cv::Point2f& axesSize, const char colorLabel)
	{
		int size_last, size_now;
		double pass_rate = 0;		//环形扫描中的通过率，指在当前状态下所有合格点所占的比例
		double add_rate  = 0;		//环形扫描中的添加率，指在每一次环形扫描中合格点所占的比例
		
		std::vector< cv::Point > traversal_points;				//环形遍历得到的像素点(栅格点)
		std::vector< cv::Point > regional_points;				//拓扑区域内所有合格点
		std::vector< cv::Point > abandoned_points;			//每次环形遍历丢弃的点
		std::vector< cv::Point > hull_points;					//凸包
		regional_points.clear();
		abandoned_points.clear();
		
		bool regrowth_flag = false;
		int growthsize_max = round(Radius_max / (maTopoMap->getResolution()));				//生长半径上限值
		int growth_size = round(axesSize.y);
		int radius = round(axesSize.x);
		if(radius != 1)					//重新生长标识
			 regrowth_flag = true;

		while(ros::ok() && radius < growth_size)
		{
			pixelSpiralIterator(pixelCenter, radius, traversal_points);			//环形遍历像素点，一次获得一个环形数据
			radius++;
			size_last = traversal_points.size();
			if(size_last == 0)
				continue;
			
			//在每一次环形遍历中，剔除不合格点
			std::vector<cv::Point>::iterator it;
			for( it = traversal_points.begin() ; it != traversal_points.end() ;  )
			{		
				//首先，筛选去除简单的不合格点
				if(checkPointISQualified( *it , colorLabel, 0) == false)
				{
					abandoned_points.push_back(*it);
					it = traversal_points.erase( it ); 
					continue;
				}
				
				//然后，判断与中心点形成的射线线是否穿过不合格区域
				std::vector< cv::Point > check_points;
				int i = 0 ;
				ImprovedBresenham(pixelCenter, *it, check_points);		//利用Bresenham划线算法计算途径的栅格点坐标
				for(i = 0 ; i < check_points.size() ; i++)
				{
					if(checkPointISQualified( check_points[i], colorLabel, 0) == false)
					{
						abandoned_points.push_back(*it);
						it = traversal_points.erase( it ); 
						break;
					}
				}
				if(i < check_points.size())
					continue;
				
				
				//条件4，检查每一个点对整体凸包性质的影响，若造成凹陷率超过某一阈值则进行删除
				regional_points.push_back( *it );						//将当前待判断点加入			
				if(regional_points.size() < 5)							//小数量的点不进行凸包计算，加快计算
					continue;
				
				//计算当前区域的凸包
// 				ConvexHull_QuickHull(regional_points, hull_points);
				ConvexHull_GrahamScanning(regional_points, hull_points);
				
				//统计目前凸包内所有的点，对目前为止的所有环状点进行筛选(每一次的凸包改变会带来越来越多的不合格点)
				//计算合格点数与不合格点数
				//从废弃池中筛选在当前凸包内的点，也就是不合格点数
				int failedpoints_size = 0 ;
				for(int j = 0 ; j < abandoned_points.size() ; j++)
				{
					if(checkPointInHull(abandoned_points[j], hull_points) == true)
						failedpoints_size++;
				}
				
				//计算总点数：原来的点+废弃池中合格的点
				int allpoints_size = regional_points.size() + failedpoints_size;
				
				//计算目前区域的障碍物比率（不合格率），若超过某一个阈值则删除该点
				if(((double)failedpoints_size/(double)allpoints_size) > ObstacleRatio)
				{
					abandoned_points.push_back(*it);
					it = traversal_points.erase( it ); 
					regional_points.pop_back();
				}else
					++it;
			}
			//计算区域占有率(点合格率)，超过某一阈值则增大生长半径
			int allpoints_size = regional_points.size() + abandoned_points.size();
			pass_rate = ((double)regional_points.size()) / (double)allpoints_size ;
// 			if(traversal_points.empty())				//无点添加时，就可以直接退出了
// 				break;
			
			//将合格点添加到当前邻域中，添加颜色标识
			for(int j = 0 ; j < traversal_points.size() ; j++ )
				maTopoMap->setTopoMapLabel(traversal_points[j], colorLabel);
			
			//发布实时消息
			maTopoMap->updateMapTime();
			topomap_pub.publish( *(maTopoMap->getTopoMap()) );
			
			size_now = traversal_points.size();
			if(regrowth_flag == false)
			{//半径动态修改
				//计算每次的添加率
				add_rate = (double)size_now / (double)size_last;
				if(add_rate > AddRatio)
				{		
					if(  pass_rate < 0.2 )
						break;
					else if(  pass_rate < PassRatio )
						growth_size = growth_size + Radius_w1 *(1 - pass_rate) * growth_size;
					
					if(radius == growthsize_max)		//如果达到半径最大阈值，直接退出生长过程
						break;
					if(growth_size > growthsize_max)		//进行半径限位控制，防止半径过大
						growth_size = growthsize_max;
				}else{//添加率太小(说明此时区域生长已经变形)，直接退出生长过程
					if(debug_flag)
						std::cout<<"mapping: 添加率太小，直接退出生长："<<add_rate<<std::endl;
					break;
				}
			}
 		}
 		if(debug_flag)
			std::cout<<"mapping: 生长结束，目前合格率为："<<pass_rate<<std::endl;
		
 		//结束生长，返回是否重调整中心点标志
		double axe_long, axe_short;
		if( (size_last - size_now) > 0 &&( !regional_points.empty()))
		{//此时说明剔除了一些点，邻域呈现不规则形状，需要调整中心点(只在遍历的最后进行调整，否则容易导致无限生长)
			cv::Point old_center = pixelCenter;
			pixelCenter = computeCenterOfGravity(old_center, regional_points);			//计算多边形的重心	
			computeEllipseAxes(regional_points, &axe_long, &axe_short);
			if(debug_flag)
				std::cout<<"mapping: 已经用区域重心当做拓扑顶点"<<(old_center)<<">>"<<(pixelCenter)<<std::endl;
		}else{
			axe_long = radius;
			axe_short = radius;
		}
		axesSize.x = axe_short;
		axesSize.y = axe_long;
		
		return true;
	}
	bool mapping::RegionalRegrowth()
	{
		cv::Point pixel_center;
		int radius_max = round(Radius_max/ (maTopoMap->getResolution()));

		std::vector<VertexNode>::iterator iter_v;
		iter_v = maTopoMap->getVertexIter();

		if((*iter_v).data.axes_size.y >= radius_max)
			return true;
		
		if(debug_flag)
			std::cout<<"RegionalRegrowth: 区域正在重新生长"<<((*iter_v).index)<<"号区域......"<<std::endl;
		cv::Point2f axes_size = cv::Point2f(1, radius_max);
		pixel_center = (*iter_v).data.center;
		axes_size.x =  ((*iter_v).data.axes_size.x)*0.5 ; 
		

		pixelDynamicGrowth(pixel_center, axes_size,  (*iter_v).data.rgb_label);
		
		(*iter_v).data.center = pixel_center;
		(*iter_v).data.axes_size = axes_size;
		addAdjEdges((*iter_v).index);

		if(maTopoMap->vertexIterIsEnd())
			return false;
		else return true;

	}
	bool mapping::RegionalMerger(const int id, const double area_max, double area_min)
	{
		//计算需要融合的顶点序号控制区域的凸包合格率，满足阈值则进行融合
		std::vector<VertexNode>::iterator iter_v;
		std::vector<int> fusion_vertexs;			//合并区域顶点序号集
		
		double ob_ratio;						//障碍物比率
		cv::Point region_center;				//合并区域的加权中心点
		VertexNodeInfo first_vertex;			//主顶点信息
		if(area_min == 0)
		{
			int radius_min = round(Radius_min / (maTopoMap->getResolution()));
			area_min = M_PI * pow(radius_min, 2) * 1.2 ;
		}

		std::vector<VertexNode>* vertexs_list = maTopoMap->getVertexsList();

		VertexNode* vertex_temp = &(*vertexs_list)[id];
		fusion_vertexs.clear();
		
		//搜索合格的合并区域，以顶点为中心，检查临近区域是否符合条件
		first_vertex =vertex_temp->data;
		double regions_area = M_PI * (first_vertex.axes_size.x) * (first_vertex.axes_size.y);
		if( (vertex_temp->edges == 1 && regions_area > area_min ) 			//末端小区域顶点区域不予考虑
			|| ( vertex_temp->edges > 1 && regions_area < area_max ) )			
		{
			fusion_vertexs.push_back(vertex_temp->index);
			for(int j = 0 ; j < vertex_temp->edges ; j++)
			{
				int adj_id = vertex_temp->adjEdges[j].adjvex;
				VertexNode* temp =  (maTopoMap->getVertexNode(adj_id));
				//FIXME 对合并区域增加限制条件: 末端节点若是小区域的话，不进行合并
				double area = M_PI * (temp->data.axes_size.x) * (temp->data.axes_size.y);
				if(area < area_min && (temp->edges) == 1)			//末端小区域顶点区域不予考虑
					continue;
				regions_area = regions_area + area;
				if(regions_area < area_max)
				{
					fusion_vertexs.push_back(adj_id);
					if(executeRegionsMerger(fusion_vertexs) == false)
					{
						fusion_vertexs.pop_back();
						regions_area = regions_area - area;
					}else break;
				}
			}
		}
		
		return true;
	}
	bool mapping::RegionalMerger2(double ratio)
	{
		std::vector<int> fusion_vertexs;			//合并区域顶点序号集
		
		std::vector<VertexNode>* vertexs_list = maTopoMap->getVertexsList();
		for(int id = 0 ; id < maTopoMap->vertex_size() ; id++)
		{
			VertexNode* vertex_temp = &(*vertexs_list)[id];
			fusion_vertexs.clear();
			fusion_vertexs.push_back(vertex_temp->index);
			for(int j = 0 ; j < vertex_temp->edges ; j++)
			{
				int id = vertex_temp->adjEdges[j].adjvex;
				fusion_vertexs.push_back(id);
				if(executeRegionsMerger(fusion_vertexs, ratio) == false)
				{
					fusion_vertexs.pop_back();
				}else break;
			}
		}
		if(debug_flag)
			std::cout<<"mapping:  融合过程结束, ratio = "<<(ratio)<<std::endl;
	}
	bool mapping::executeRegionsMerger(std::vector< int >& gather, double ratio)
	{
		double ob_ratio;							//障碍物比率
		cv::Point region_center;					//合并区域的加权中心点
		std::vector< cv::Point > fusion_points;		//融合区域内所有的点
		std::vector< cv::Point > ring_points;		//环形遍历得到的像素点(栅格点)
		std::vector< cv::Point > failed_points;		//环形遍历过程中的不合格点
		std::vector< cv::Point > hull_points;					//凸包
		char new_label;
		
		double ratio_ = ratio;
		if( ratio == 0)
			ratio_ = ObstacleRatio*1.3;
		ob_ratio = ratio_;
		//计算待合并区域的障碍物比率
		if(gather.size() >= 2 )		//顶点数为2个或以上
		{
			
			//取加权中心点和半径
			VertexNode* vertex1 = maTopoMap->getVertexNode(gather[0]);
			VertexNode* vertex2 = maTopoMap->getVertexNode(gather[1]);
			region_center = ( (vertex1->data.center + vertex2->data.center) / 2.0);
			int region_range = round((vertex1->data.axes_size.y + vertex1->data.axes_size.y) * 1.5);
			new_label = vertex1->data.rgb_label;
			
			//pthread_rwlock_unlock(&(topomap_lock));      //写者释放写锁
			fusion_points.clear();
			failed_points.clear();
			bool scan_flag = true;
			int range = 1;
			while(ros::ok() && scan_flag == true && range < region_range)
			{
				scan_flag = pixelSpiralIterator(region_center, range, ring_points);	//环形遍历像素点，一次获得一个环形数据
				range++;
				
				for(int j = 0 ; j < ring_points.size() ; j++)
				{
					if(checkPointISQualified(ring_points[j], vertex1->data.rgb_label, vertex2->data.rgb_label) == true)
						fusion_points.push_back(ring_points[j]);
					else
						failed_points.push_back(ring_points[j]);
				}
				if(ring_points.empty())
					break;
			}
			if(fusion_points.empty() == false)
			{//计算当前区域的凸包
		// 		ConvexHull_QuickHull(fusion_points, hull_points);
				ConvexHull_GrahamScanning(fusion_points, hull_points);
				
				int failedpoints_size = 0 ;
				for(int i = 0 ; i < failed_points.size() ; i++)
				{
					if(checkPointInHull(failed_points[i], hull_points) == true)
						failedpoints_size++;
				}
				
				//计算总点数：原来的点+废弃池中合格的点
				int allpoints_size = fusion_points.size() + failedpoints_size;
				
				//计算目前区域的障碍物比率（不合格率），若小于某一个阈值则符合融合条件
				ob_ratio = ((double)failedpoints_size/(double)allpoints_size);
					
			}else return false;
		}else return false;
			
		
		//检查障碍物比率，符合条件即进行合并
		if( ob_ratio < ratio_)
		{
			if(debug_flag)
				std::cout<<"RegionalMerger : 区域"<<(gather[0])<<"与"<<(gather[1])<<"马上进行合并！ "<<std::endl;

			//删除即将要融合的顶点
			  for(int i = 0 ; i < gather.size() ; i++)
				  maTopoMap->deleteAVertex(gather[i], true);
			  
			  //进行融合，对区域标识进行替换
			  for(int i = 0 ; i < fusion_points.size() ; i++)
				  maTopoMap->setTopoMapLabel(fusion_points[i], new_label);
			  maTopoMap->updateMapTime();
			  topomap_pub.publish( *(maTopoMap->getTopoMap()) );
			  
			  //建立新节点
			  cv::Point center_v = computeCenterOfGravity(region_center, fusion_points);
			  double longAxe, shortAxe;
			  computeEllipseAxes(fusion_points, &longAxe, &shortAxe);
			  establishATopoVertex(cv::Point2f(shortAxe, longAxe), center_v, new_label);
			  
			  return true;
		}else{
			if(debug_flag)
				std::cout<<"RegionalMerger : 区域"<<(gather[0])<<"与"<<(gather[1])<<"合并, 障碍物比率过大 "<<(ob_ratio)<<std::endl;
			return false;
		}
	}
	void mapping::NeighborhoodSegmentation(cv::Point2f pose)
	{
		static cv::Point2f growth_point = cv::Point2f(0, 0);
		cv::Point2f axes_size = cv::Point2f(1, round(Radius_min / (maTopoMap->getResolution())));
		

		//初始生长点选择
// 		if(chooseBestGrowthPoint(growth_point))
// 		{
// 			cv::Point pixel_center = maTopoMap->map2pixelCoord(growth_point, true);
// 			pixelDynamicGrowth(pixel_center, axes_size, Pen_color);		//区域生长，生长结束后才能退出子函数
// 			establishATopoVertex(axes_size, pixel_center);
// 			changePenColor();									//区域颜色标签更换
// 			//growth_flag = false;	//初始化生长状态标志位
// 		}else{//生长点未找到，生长未开始，更新一些状态
// 			//GrowthStatusUpdate();
// 		}

		if(checkBestGrowthPoint(pose))
		{
			cv::Point pixel_center = maTopoMap->map2pixelCoord(pose, true);
			pixelDynamicGrowth(pixel_center, axes_size, Pen_color);		//区域生长，生长结束后才能退出子函数
			establishATopoVertex(axes_size, pixel_center);
			changePenColor();									//区域颜色标签更换
		}

	}
	bool mapping::checkPointISQualified(const cv::Point p, const char colorLabel1, const char colorLabel2, char occValue)
	{//点p的坐标是topomap中的坐标
		 double map_ratio =  maTopoMap->getResolution() / maTopoMap->getResolution(false);
		
		//若该点为障碍物点或者未知点，则认为是不合格点
		if(maTopoMap->getOccValue( p * map_ratio) != occValue)
			return false;
		
		int label = maTopoMap->getTopoLabel( p );
		if(colorLabel1 != colorLabel2 || colorLabel1 != 0)
		{//若该点满足邻域属性同时不为label1和label2的条件，则认为是不合格点
			if(label != colorLabel1 && label != colorLabel2)
				return false;
		}else{//若该点领域属性为label1(label1=label2)，则认为是不合格点
			if(label == colorLabel1)
				return false;
		}
		
		return true;
	}

	





}
