/*
 * This file contains helper functions for loading images as maps.
 *
 * Author: Liu Yuqiang
 */

#include <cstring>
#include <stdexcept>

#include <stdlib.h>
#include <stdio.h>
#include <algorithm>

// We use SDL_image to load the image from disk
#include <SDL/SDL_image.h>

#include "topomap_server/image_loader.h"
#include <tf/tf.h>

// compute linear index for given map coords
#define MAP_IDX(sx, i, j) ((sx) * (j) + (i))

namespace topomap_server
{

void loadMapFromFile(nav_msgs::GetMap::Response* gridmap_resp,
                const char* fname, double res, bool negate,
                double occ_th, double free_th, double* origin,
                TopoMapMode mode)
{
	SDL_Surface* img;

	unsigned char* pixels;
	unsigned char* p;
	unsigned char value;
	int rowstride, n_channels, avg_channels;
	unsigned int i,j;
	int k;
	double occ;
	int alpha;
	int color_sum;
	double color_avg;

	// Load the image using SDL.  If we get NULL back, the image load failed.
	if(!(img = IMG_Load(fname)))
	{
		std::string errmsg = std::string("failed to open image file \"") +
		std::string(fname) + std::string("\"");
		throw std::runtime_error(errmsg);
	}

	// Copy the image data into the map structure
	gridmap_resp->map.info.width = img->w;
	gridmap_resp->map.info.height = img->h;
	gridmap_resp->map.info.resolution = res;
	gridmap_resp->map.info.origin.position.x = *(origin);
	gridmap_resp->map.info.origin.position.y = *(origin+1);
	gridmap_resp->map.info.origin.position.z = 0.0;
	tf::Quaternion q;
	q.setRPY(0,0, *(origin+2));
	gridmap_resp->map.info.origin.orientation.x = q.x();
	gridmap_resp->map.info.origin.orientation.y = q.y();
	gridmap_resp->map.info.origin.orientation.z = q.z();
	gridmap_resp->map.info.origin.orientation.w = q.w();

	// Allocate space to hold the data
	gridmap_resp->map.data.resize(gridmap_resp->map.info.width * gridmap_resp->map.info.height);

	// Get values that we'll need to iterate through the pixels
	rowstride = img->pitch;
	n_channels = img->format->BytesPerPixel;

	// NOTE: Trinary mode still overrides here to preserve existing behavior.
	// Alpha will be averaged in with color channels when using trinary mode.
	if (mode==TRINARY || !img->format->Amask)
		avg_channels = n_channels;
	else
		avg_channels = n_channels - 1;

	// Copy pixel data into the map structure
	pixels = (unsigned char*)(img->pixels);
	for(j = 0; j < gridmap_resp->map.info.height; j++)
	{
		for (i = 0; i < gridmap_resp->map.info.width; i++)
		{
			// Compute mean of RGB for this pixel
			p = pixels + j*rowstride + i*n_channels;
			color_sum = 0;
			for(k=0;k<avg_channels;k++)
				color_sum += *(p + (k));
			color_avg = color_sum / (double)avg_channels;

			if (n_channels == 1)
				alpha = 1;
			else
				alpha = *(p+n_channels-1);

			if(negate)
				color_avg = 255 - color_avg;

			if(mode==RAW)
			{
				value = color_avg;
				gridmap_resp->map.data[MAP_IDX(gridmap_resp->map.info.width, i, gridmap_resp->map.info.height - j - 1)] = value;
				continue;
			}

			// If negate is true, we consider blacker pixels free, and whiter
			// pixels free.  Otherwise, it's vice versa.
			occ = (255 - color_avg) / 255.0;

			// Apply thresholds to RGB means to determine occupancy values for
			// map.  Note that we invert the graphics-ordering of the pixels to
			// produce a map with cell (0,0) in the lower-left corner.
			if(occ > occ_th)			//占有
				value = +100;
			else if(occ < free_th)		//空闲
				value = 0;
			else if(mode==TRINARY || alpha < 1.0)
				value = -1;			//未知
			else {
				double ratio = (occ - free_th) / (occ_th - free_th);
				value = 99 * ratio;
			}

			gridmap_resp->map.data[MAP_IDX(gridmap_resp->map.info.width, i, gridmap_resp->map.info.height - j - 1)] = value;
		}
	}
	SDL_FreeSurface(img);//用来删除RGB表面
}

void RGBtoHSV( float r, float g, float b, float *h, float *s, float *v )
{
    float min, max, delta;
    min = std::min( r, std::min( g, b ));
    max = std::max( r, std::max( g, b ));
    *v = max;               // v
    delta = max - min;
    if( max != 0 )
        *s = delta / max;       // s
    else {
        // r = g = b = 0        // s = 0, v is undefined
        *s = 0;
        *h = -1;
        return;
    }
    if( r == max )
        *h = ( g - b ) / delta;     // between yellow & magenta
    else if( g == max )
        *h = 2 + ( b - r ) / delta; // between cyan & yellow
    else
        *h = 4 + ( r - g ) / delta; // between magenta & cyan
    *h *= 60;               // degrees
    if( *h < 0 )
        *h += 360;
}
//float h /* 0~360 degrees */, float s /* 0 ~ 1.0 */, float v /* 0 ~ 1.0 */
void HSVtoRGB(float h, float s, float v,  float *r, float *g, float *b )
{
    int i;
    float f, p, q, t;
    if( s == 0 ) {
        // achromatic (grey)
        *r = *g = *b = v;
        return;
    }
    h /= 60;            // sector 0 to 5
    i = floor( h );
    f = h - i;          // factorial part of h
    p = v * ( 1 - s );
    q = v * ( 1 - s * f );
    t = v * ( 1 - s * ( 1 - f ) );
    switch( i ) {
        case 0:
            *r = v;
            *g = t;
            *b = p;
            break;
        case 1:
            *r = q;
            *g = v;
            *b = p;
            break;
        case 2:
            *r = p;
            *g = v;
            *b = t;
            break;
        case 3:
            *r = p;
            *g = q;
            *b = v;
            break;
        case 4:
            *r = t;
            *g = p;
            *b = v;
            break;
        default:        // case 5:
            *r = v;
            *g = p;
            *b = q;
            break;
    }
}






}
