#include <stdio.h>
#include <iostream>
#include <cstdio>
#include <boost/graph/graph_concepts.hpp>
#include "ros/ros.h"
#include "ros/console.h"
#include "nav_msgs/GetMap.h"
#include "tf/LinearMath/Matrix3x3.h"
#include "geometry_msgs/Quaternion.h"
#include "geometry_msgs/Point32.h"
#include "topo_mapping/vertex.h"

#define MAX_SIZE 50

using namespace std;

/**
 * @brief Map generation node.
 * 文件描述子		类型		编码
 * P1			位图		ASCII
 * P2			灰度图	ASCII
 * P3			像素图	ASCII
 * P4			位图		二进制
 * P5			灰度图	二进制
 * P6			像素图	二进制
 */
class TopoMapGenerator
{
	typedef struct AreaLabel
	{
		unsigned char r;
		unsigned char g;
		unsigned char b;
	}AreaLabel;
	
	public:
	TopoMapGenerator(const std::string& mapname) : mapname_(mapname), saved_map_(false), rece_topomap_(false)
	{
		ros::NodeHandle n;
		ROS_INFO("Waiting for the map");
		gridmap_sub_ = n.subscribe("map", 1, &TopoMapGenerator::gridmapCallback, this);
		topomap_sub_ = n.subscribe("/topomap", 1, &TopoMapGenerator::topomapCallback, this);
		
		rgb_labels[0].r = 255; rgb_labels[0].g = 0; rgb_labels[0].b = 255;//品红
		rgb_labels[1].r = 255; rgb_labels[1].g = 192; rgb_labels[1].b = 203;//粉
		rgb_labels[2].r = 255; rgb_labels[2].g = 99; rgb_labels[2].b = 71;//番茄
		
		rgb_labels[3].r = 255; rgb_labels[3].g = 165; rgb_labels[3].b = 0;//橙子
		rgb_labels[4].r = 250; rgb_labels[4].g = 128; rgb_labels[4].b = 114;//三文鱼
		rgb_labels[5].r = 210; rgb_labels[5].g = 105; rgb_labels[5].b = 30;//巧克力
		
		rgb_labels[6].r = 0; rgb_labels[6].g = 0; rgb_labels[6].b = 0;//黑色，表示拓扑结点位置
		
		rgb_labels[7].r = 147; rgb_labels[7].g = 112; rgb_labels[7].b = 219;//暗紫
		rgb_labels[8].r = 218; rgb_labels[8].g = 112; rgb_labels[8].b = 214;//兰花
		rgb_labels[9].r = 221; rgb_labels[9].g = 160; rgb_labels[9].b = 221;//李子
		
		rgb_labels[10].r = 255; rgb_labels[10].g = 215; rgb_labels[10].b = 0;//金
		rgb_labels[11].r = 189; rgb_labels[11].g = 183; rgb_labels[11].b = 107;//深卡其
		rgb_labels[12].r = 205; rgb_labels[12].g = 133; rgb_labels[12].b = 63;//秘鲁
	}

	void gridmapCallback(const nav_msgs::OccupancyGridConstPtr& gridmap)
	{
		ROS_INFO("Received a %d X %d gridmap @ %.3f m/pix",
				gridmap->info.width,
				gridmap->info.height,
				gridmap->info.resolution);
		
		while( ros::ok() && (rece_topomap_ == false) )
		{
			ROS_INFO_STREAM_ONCE("Waiting for the topomap ......");
			ros::spinOnce();
		}
		
		
		std::string mapdatafile = mapname_ + ".ppm";
		ROS_INFO("Writing topomap label data to %s", mapdatafile.c_str());
		FILE* out = fopen(mapdatafile.c_str(), "w");
		if (!out)
		{
			ROS_ERROR("Couldn't save topomap file to %s", mapdatafile.c_str());
			return;
		}

		fprintf(out, "P6\n# CREATOR: topomap_saver.cpp %.3f m/pix\n%d %d\n255\n",gridmap->info.resolution, gridmap->info.width, gridmap->info.height);
		//char *result;
		char label;
		unsigned char pen_;
		for(unsigned int y = 0; y < gridmap->info.height; y++)
		{
			for(unsigned int x = 0; x < gridmap->info.width; x++) 
			{
				unsigned int i = x + (gridmap->info.height - y - 1) * gridmap->info.width;
				if (gridmap->data[i] == 0) 
				{ //occ [0,0.1)表示空闲区域
					//fputc(254, out);
					label = topomap_->data[i];
					if(label != 0)
					{//拓扑结构区域，按照区域属性填充颜色
// 						if(label == +127)
// 							pen_ = 6;						//绘制拓扑节点位置
// 						else
							pen_ = abs((label + 120)/20);	//绘制拓扑区域
						
						fputc(rgb_labels[pen_].r, out);		//R
						fputc(rgb_labels[pen_].g, out);		//G
						fputc(rgb_labels[pen_].b, out);		//B
						
					}else{//非拓扑结构区域，默认绿色
						fputc(0, out);		//R
						fputc(254, out);	//G
						fputc(0, out);		//B
					}
				} else if (gridmap->data[i] == +100) 
				{ //occ (0.65,1]表示障碍区域
					//fputc(000, out);
					fputc(254, out);	//R
					fputc(0, out);		//G
					fputc(0, out);		//B
				} else { //occ [0.1,0.65]表示未知空间
					//fputc(205, out);
					fputc(0, out);		//R
					fputc(0, out);		//G
					fputc(254, out);	//B
				}
			}
		}

		fclose(out);

		std::string mapmetadatafile = mapname_ + ".yaml";
		ROS_INFO("Writing topomap message data to %s", mapmetadatafile.c_str());
		FILE* yaml = fopen(mapmetadatafile.c_str(), "w");
/*
resolution: 0.100000
origin: [0.000000, 0.000000, 0.000000]
#
negate: 0
occupied_thresh: 0.65
free_thresh: 0.196*/

		geometry_msgs::Quaternion orientation = gridmap->info.origin.orientation;
		tf::Matrix3x3 mat(tf::Quaternion(orientation.x, orientation.y, orientation.z, orientation.w));
		double yaw, pitch, roll;
		mat.getEulerYPR(yaw, pitch, roll);

		fprintf(yaml, "image: %s\nresolution: %f\norigin: [%f, %f, %f]\nnegate: 0\noccupied_thresh: 0.65\nfree_thresh: 0.196\n\n",
		mapdatafile.c_str(), gridmap->info.resolution, gridmap->info.origin.position.x, gridmap->info.origin.position.y, yaw);

		
		//请求拓扑顶点信息
		ros::ServiceClient vertex_client = private_nh.serviceClient<topo_mapping::vertex>("/dynamic_topomap");
		topo_mapping::vertex vertex_srv;
		vertex_srv.request.order = 1;
		if(vertex_client.call(vertex_srv) == true)
		{
			int vertex_len = vertex_srv.response.vertex.size();
			fprintf(yaml, "vertexs_num: %i\n", vertex_len/4);
			for(int i = 0 ; i < vertex_len ; i = i + 4)
			{
				fprintf(yaml, "%f %f %f %f\n", vertex_srv.response.vertex[i], vertex_srv.response.vertex[i+1], 
						vertex_srv.response.vertex[i+2], vertex_srv.response.vertex[i+3]);	
			}
		}else{
			ROS_ERROR("Failed to call vertex service, request topology vertex information failed !");
			//return;
		}
		vertex_srv.request.order = 2;
		vertex_srv.response.vertex.clear();
		if(vertex_client.call(vertex_srv) == true)
		{
			fprintf(yaml, "obstacle_ratio: %f\n", vertex_srv.response.vertex[0]);	
			fprintf(yaml, "add_ratio: %f\n", vertex_srv.response.vertex[1]);	
			fprintf(yaml, "pass_ratio: %f\n", vertex_srv.response.vertex[2]);	
			fprintf(yaml, "radius_weight1: %f\n", vertex_srv.response.vertex[3]);	
		}else{
			ROS_ERROR("Failed to call vertex service, request topology vertex information failed !");
			//return;
		}
		
		
		fclose(yaml);

		ROS_INFO("Done\n");
		saved_map_ = true;
	}

     void topomapCallback(const nav_msgs::OccupancyGridConstPtr& topomap)
     {
	      ROS_INFO("Received a %d X %d topomap @ %.3f m/pix",
               topomap->info.width,
               topomap->info.height,
               topomap->info.resolution);
	      
	      topomap_ = topomap;
	      rece_topomap_ = true;
	}
    
	ros::NodeHandle private_nh;
	std::string mapname_;
	ros::Subscriber gridmap_sub_, topomap_sub_;
	bool saved_map_, rece_topomap_;
	nav_msgs::OccupancyGridConstPtr topomap_;
	AreaLabel rgb_labels[13];
};

#define USAGE "Usage: \n" \
              "  map_saver -h\n"\
              "  map_saver [-f <mapname>] [ROS remapping args]"

int main(int argc, char** argv)
{
	ros::init(argc, argv, "map_saver");
	std::string mapname = "map";

	for(int i=1; i<argc; i++)
	{
		if(!strcmp(argv[i], "-h"))
		{
			puts(USAGE);
			return 0;
		}else if(!strcmp(argv[i], "-f"))
		{
			if(++i < argc)
				mapname = argv[i];
			else{
				puts(USAGE);
				return 1;
			}
		}else{
			puts(USAGE);
			return 1;
		}
	}

	TopoMapGenerator mg(mapname);

	while(!mg.saved_map_ && ros::ok())
		ros::spinOnce();

	return 0;
}


